(* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *  Copyright (C) 2004, 2005 Milenko "DCoder" Mitrovic                       *
 *  Mail: dcoder@dsp-worx.de                                                 *
 *  Web:  http://www.dsp-worx.de                                             *
 *                                                                           *
 *  This Program is free software; you can redistribute it and/or modify     *
 *  it under the terms of the GNU General Public License as published by     *
 *  the Free Software Foundation; either version 2, or (at your option)      *
 *  any later version.                                                       *
 *                                                                           *
 *  This Program is distributed in the hope that it will be useful,          *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the             *
 *  GNU General Public License for more details.                             *
 *                                                                           *
 *  You should have received a copy of the GNU General Public License        *
 *  along with GNU Make; see the file COPYING.  If not, write to             *
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.    *
 *  http://www.gnu.org/copyleft/gpl.html                                     *
 *                                                                           *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *)

unit DSoundRenderer;

interface

uses
  BaseClass, DirectSound, ActiveX, Math, Windows, SysUtils, Classes,
  DirectShow9, MMSystem, DSUtil, Registry, DSFading, Messages, PropertyPages;

const
  CLSID_CrossfadingDirectSoundRenderer: TGUID = '{C70BD94C-ABB7-409B-B95A-EC0285FF3D8C}';

  SudPinTypes : array[0..5] of TRegPinTypes =
  (
    (clsMajorType: @MEDIATYPE_AUDIO;clsMinorType: @MEDIASUBTYPE_PCM),
    (clsMajorType: @MEDIATYPE_AUDIO;clsMinorType: @MEDIASUBTYPE_IEEE_FLOAT),
    (clsMajorType: @MEDIATYPE_AUDIO;clsMinorType: @MEDIASUBTYPE_DOLBY_AC3_SPDIF),
    (clsMajorType: @MEDIATYPE_AUDIO;clsMinorType: @MEDIASUBTYPE_RAW_SPORT),
    (clsMajorType: @MEDIATYPE_AUDIO;clsMinorType: @MEDIASUBTYPE_SPDIF_TAG_241h),
    (clsMajorType: @MEDIATYPE_AUDIO;clsMinorType: @MEDIASUBTYPE_DRM_Audio)
  );

  SudPins : array[0..0] of TRegFilterPins =
  (
    (strName: 'Input'; bRendered: TRUE; bOutput: FALSE; bZero: FALSE; bMany: FALSE; oFilter: nil; strConnectsToPin: nil; nMediaTypes: 6; lpMediaType: @SudPinTypes)
  );

type
  TCrossfadingDirectSoundRenderer = class;

  TMyRendererPosPassThru = class(TBCRendererPosPassThru, IMediaSeeking)
  private
    FOwner: TCrossfadingDirectSoundRenderer;
    FLastSeek: Cardinal;
    FLastPos: Int64;
  public
    constructor Create(Owner: TCrossfadingDirectSoundRenderer; name: String; Unk: IUnknown; out hr: HRESULT; Pin: IPin); reintroduce;
    function SetRate(dRate: Double): HRESULT; override; stdcall;
    function put_Rate(dRate: Double): HRESULT; override; stdcall;
    function SetPositions(var pCurrent: Int64; dwCurrentFlags: Cardinal;
      var pStop: Int64; dwStopFlags: Cardinal): HRESULT; override; stdcall;
  end;

  TCrossfadingDirectSoundRenderer = class(TBCBaseRenderer, IBasicAudio,
                                          IPersist, IMediaFilter, ISpecifyPropertyPages,
                                          ICrossfadingRenderer, IAMFilterMiscFlags, IDispatch)
  private
    FRate: Single;
    FGuid: TGuid;
    FFadingSeek: Integer;
    FFadingCross: Integer;
    FAllowVideo: Boolean;
    FLastFade: Integer;

    FCurrentFadingSeek: Integer;
    FCurrentFadingCross: Integer;

    FFirstRun: Boolean;
    FPlayThread: TDSPlayThread;
    FDispatch: TBCBaseDispatch;
    FFormat: TWaveFormatExtensible;
    FCurrentBufferSize: DWORD;
    FPreBufferSize: DWORD;
    FRefClock: TBCSystemClock;
    FVolume: integer;
    FBalance: integer;
    FBuffer: IDirectSoundBuffer;
    FPos: DWORD;
    FBufferSize : DWORD;
    FMediaType: PAMMediaType;
    (*** Misc methods ***)
    procedure FadeIn;
    procedure FadeOut(Seek: Boolean);
    function CreateDirectSoundBuffer(MediaType : PAMMediaType) : HRESULT;
    procedure KillFading;
    procedure LoadSettings;
    procedure SaveSettings;
    function CheckVideoRenderer: Boolean;
    procedure SetRate(dRate: Double);
  public
    (*** TBCBaseRenderer methods ***)
    constructor Create(ObjName: String; Unk: IUnknown; out hr : HResult);
    constructor CreateFromFactory(Factory: TBCClassFactory; const Controller: IUnknown); override;
    destructor Destroy; override;
    function CheckMediaType(MediaType: PAMMediaType): HResult; override;
    function DoRenderSample(MediaSample: IMediaSample): HResult; override;
    function SetMediaType(MediaType: PAMMediaType): HResult; override;
    function Run(StartTime: TReferenceTime): HResult; override; stdcall;
    function Stop: HResult; override; stdcall;
    function Pause: HResult; override; stdcall;
    function Receive(MediaSample: IMediaSample): HResult; override;
    function BeginFlush: HResult; override;
    function EndOfStream: HResult; override;
    function NonDelegatingQueryInterface(const IID: TGUID; out Obj): HResult; override; stdcall;
    function GetMediaPositionInterface(IID: TGUID; out Obj): HResult;  override;
    (*** IDispatch methods ***)
    function GetTypeInfoCount(out Count: Integer): HResult; stdcall;
    function GetTypeInfo(Index, LocaleID: Integer; out TypeInfo): HResult; stdcall;
    function GetIDsOfNames(const IID: TGUID; Names: Pointer; NameCount, LocaleID: Integer; DispIDs: Pointer): HResult; stdcall;
    function Invoke(DispID: Integer; const IID: TGUID; LocaleID: Integer; Flags: Word; var Params; VarResult, ExcepInfo, ArgErr: Pointer): HResult; stdcall;
    (*** IBasicAudio methods ***)
    function put_Volume(lVolume: Longint): HResult; stdcall;
    function get_Volume(out plVolume: Longint): HResult; stdcall;
    function put_Balance(lBalance: Longint): HResult; stdcall;
    function get_Balance(out plBalance: Longint): HResult; stdcall;
    (*** IReferenceClock methods ***)
    function GetTime(out pTime: int64): HResult; stdcall;
    function AdviseTime(rtBaseTime, rtStreamTime: int64; hEvent: THandle; out pdwAdviseCookie: DWORD): HResult; stdcall;
    function AdvisePeriodic(const rtStartTime, rtPeriodTime: int64; hSemaphore: THandle; out pdwAdviseCookie: DWORD): HResult; stdcall;
    function Unadvise(dwAdviseCookie: DWORD): HResult; stdcall;
    (*** ISpecifyPropertyPages methods ***)
    function GetPages(out pages: TCAGUID): HResult; stdcall;
    (*** ICrossfadingRenderer methods ***)
    function GetDevice(out Guid: TGuid): HRESULT; stdcall;
    function SetDevice(Guid: TGuid): HRESULT; stdcall;
    function GetFadingSeekStop(out Fading: Integer): HRESULT; stdcall;
    function SetFadingSeekStop(Fading: Integer): HRESULT; stdcall;
    function GetFadingCross(out Fading: Integer): HRESULT; stdcall;
    function SetFadingCross(Fading: Integer): HRESULT; stdcall;
    function GetAllowVideo(out Allow: Boolean): HRESULT; stdcall;
    function SetAllowVideo(Allow: Boolean): HRESULT; stdcall;
    (*** IAMFilterMiscFlags methods ***)
    function GetMiscFlags: ULONG; stdcall;
  end;

implementation

var
  LastWasEOS: Boolean = False;

constructor TMyRendererPosPassThru.Create(Owner: TCrossfadingDirectSoundRenderer; name: String; Unk: IUnknown; out hr: HRESULT; Pin: IPin);
begin
  inherited Create(name, Unk, hr, Pin);
  Pointer(FOwner) := Owner;
  FLastSeek := 0;
  FLastPos := -1;
end;

function TMyRendererPosPassThru.SetRate(dRate: Double): HRESULT;
begin
  Result := inherited SetRate(dRate);
  FOwner.SetRate(dRate);
end;

function TMyRendererPosPassThru.put_Rate(dRate: Double): HRESULT;
begin
  Result := inherited put_Rate(dRate);
  FOwner.SetRate(dRate);
end;

function TMyRendererPosPassThru.SetPositions(var pCurrent: Int64; dwCurrentFlags: Cardinal; var pStop: Int64; dwStopFlags: Cardinal): HRESULT;
var
  t1, t2: Int64;
  c: Cardinal;
begin
  c := GetTickCount;

  if FLastPos = pCurrent then
  begin
    Result := S_FALSE;
    Exit;
  end;
  FLastPos := pCurrent;

  if (c - FLastSeek <= 100) and (FOwner.State <> State_Stopped) then
  begin
    Result := S_FALSE;
    Exit;
  end;
  FLastSeek := c;

  if (FOwner.State <> State_Stopped) then
  begin
    if FOwner.FCurrentFadingSeek > FOwner.FCurrentFadingCross
      then t1 := FOwner.FCurrentFadingSeek
      else t1 := FOwner.FCurrentFadingCross;

    t1 := (t1 + 1000) * 10000;

    GetDuration(t2);
    t2 := t2 - t1;

    if (pCurrent > t2)
      then pCurrent := t2;

    if pCurrent < 0
      then pCurrent := 0;
  end;
  
  Result := inherited SetPositions(pCurrent, dwCurrentFlags, pStop, dwStopFlags);
end;

{*** IDispatch methods *** taken from CBasicAudio *** ctlutil.cpp *************}
function TCrossfadingDirectSoundRenderer.GetTypeInfoCount(out Count: Integer): HResult; stdcall;
begin
  Result := FDispatch.GetTypeInfoCount(Count);
end;

function TCrossfadingDirectSoundRenderer.GetTypeInfo(Index, LocaleID: Integer; out TypeInfo): HResult; stdcall;
begin
  Result := FDispatch.GetTypeInfo(IID_IBasicAudio, Index, LocaleID, TypeInfo);
end;

function TCrossfadingDirectSoundRenderer.GetIDsOfNames(const IID: TGUID; Names: Pointer; NameCount, LocaleID: Integer; DispIDs: Pointer): HResult; stdcall;
begin
  Result := FDispatch.GetIDsOfNames(IID_IBasicAudio, Names, NameCount, LocaleID, DispIDs);
end;

function TCrossfadingDirectSoundRenderer.Invoke(DispID: Integer; const IID: TGUID; LocaleID: Integer; Flags: Word; var Params; VarResult, ExcepInfo, ArgErr: Pointer): HResult; stdcall;
var
  pti : ITypeInfo;
begin
  if not IsEqualGUID(GUID_NULL, IID) then
  begin
    Result := DISP_E_UNKNOWNINTERFACE;
    Exit;
  end;

  Result := GetTypeInfo(0, LocaleID, pti);

  if FAILED(Result)
    then Exit;

  Result := pti.Invoke(Pointer(Self as IBasicAudio), DispID, Flags, TDispParams(Params), VarResult, ExcepInfo, ArgErr);
  pti := nil;
end;
{*** IBasicAudio methods ******************************************************}
function TCrossfadingDirectSoundRenderer.put_Volume(lVolume: Longint): HResult; stdcall;
begin
  KillFading;
  FVolume := EnsureRange(lVolume,-10000,0);
  if Assigned(FBuffer)
    then FBuffer.SetVolume(FVolume);
  Result := S_OK;
end;

function TCrossfadingDirectSoundRenderer.get_Volume(out plVolume: Longint): HResult; stdcall;
begin
  plVolume := FVolume;
  Result := S_OK;
end;

function TCrossfadingDirectSoundRenderer.put_Balance(lBalance: Longint): HResult; stdcall;
begin
  FBalance := EnsureRange(lBalance,-10000,10000);
  if Assigned(FBuffer)
    then FBuffer.SetPan(FBalance);
  Result := S_OK;
end;

function TCrossfadingDirectSoundRenderer.get_Balance(out plBalance: Longint): HResult; stdcall;
begin
  plBalance := FBalance;
  Result := S_OK;
end;
{*** IReferenceClock methods **************************************************}
function TCrossfadingDirectSoundRenderer.GetTime(out pTime: int64): HResult; stdcall;
begin
  Result := FRefClock.GetTime(pTime);
end;

function TCrossfadingDirectSoundRenderer.AdviseTime(rtBaseTime, rtStreamTime: int64; hEvent: THandle; out pdwAdviseCookie: DWORD): HResult; stdcall;
begin
  Result := FRefClock.AdviseTime(rtBaseTime, rtStreamTime, hEvent, pdwAdviseCookie);
end;

function TCrossfadingDirectSoundRenderer.AdvisePeriodic(const rtStartTime, rtPeriodTime: int64; hSemaphore: THandle; out pdwAdviseCookie: DWORD): HResult; stdcall;
begin
  Result := FRefClock.AdvisePeriodic(rtStartTime, rtPeriodTime, hSemaphore, pdwAdviseCookie);
end;

function TCrossfadingDirectSoundRenderer.Unadvise(dwAdviseCookie: DWORD): HResult; stdcall;
begin
  Result := FRefClock.Unadvise(dwAdviseCookie);
end;
{*** ISpecifyPropertyPages methods ********************************************}
function TCrossfadingDirectSoundRenderer.GetPages(out pages: TCAGUID): HResult; stdcall;
begin
  Pages.cElems := 2;
  Pages.pElems := CoTaskMemAlloc(sizeof(TGUID) * Pages.cElems);
  if (Pages.pElems = nil) then
  begin
    Result := E_OUTOFMEMORY;
    Exit;
  end;
  Pages.pElems^[0] := CLSID_PropPageSettings;
  Pages.pElems^[1] := CLSID_PropAbout;
  Result := S_OK;
end;
{*** TBCBaseRenderer methods **************************************************}
constructor TCrossfadingDirectSoundRenderer.Create(ObjName: String; Unk: IUnknown; out hr: HResult);
begin
  inherited Create(CLSID_CrossfadingDirectSoundRenderer, 'Crossfading DirectSound Renderer', Unk, hr);
  LoadSettings;
  FPlayThread := TDSPlayThread.Create(FGuid);
  while (not FPlayThread.Created) do
    sleep(1);
  if (not Assigned(FPlayThread.Device)) then
  begin
    hr := E_FAIL;
    Exit;
  end;
  FLastFade := 0;
  FRate := 1.0;
  FFirstRun := True;
  FPlayThread.Device.SetCooperativeLevel(GetDesktopWindow, DSSCL_PRIORITY);
  FDispatch := TBCBaseDispatch.Create;
  FRefClock := TBCSystemClock.Create('TBCSystemClock', nil, hr);
  FPos := 0;
  FVolume := 0;
  FBalance := 0;
  FCurrentBufferSize := 0;
  FMediaType := nil;
end;

constructor TCrossfadingDirectSoundRenderer.CreateFromFactory(Factory: TBCClassFactory;  const Controller: IUnknown);
var
  hr: HRESULT;
begin
  Create(Factory.Name, Controller, hr);
end;

destructor TCrossfadingDirectSoundRenderer.Destroy;
begin
  FBuffer := nil;

  if Assigned(FRefClock)
    then FreeAndNil(FRefClock);
  if Assigned(FDispatch)
    then FreeAndNil(FDispatch);
  if Assigned(FMediaType)
    then DeleteMediaType(FMediaType);

  if Assigned(FPlayThread)
    then PostThreadMessage(FPlayThread.ThreadID, WM_DESTROY, FLastFade, 0);
  SaveSettings;
  inherited Destroy;
end;

procedure TCrossfadingDirectSoundRenderer.LoadSettings;
var
  Reg: TRegistry;
begin
  FGuid := GUID_NULL;
  FFadingSeek := 500;
  FFadingCross := 2000;
  FAllowVideo := False;
  Reg := TRegistry.Create;
  Reg.Rootkey := HKEY_CURRENT_USER;
  if Reg.OpenKey('SOFTWARE\DSP-worx\Crossfading Renderer', False) then
  begin
    if Reg.ValueExists('Device') then
    begin
      try
        FGuid := StringToGUID(Reg.ReadString('Device'));
      except
        FGuid := GUID_NULL;
      end;
    end;

    if Reg.ValueExists('AllowVideo')
      then FAllowVideo := Reg.ReadBool('AllowVideo');

    if Reg.ValueExists('FadingSeek')
      then FFadingSeek := Reg.ReadInteger('FadingSeek');

    if Reg.ValueExists('FadingCross')
      then FFadingCross := Reg.ReadInteger('FadingCross');

    Reg.CloseKey;
  end;
  Reg.Free;

  FCurrentFadingSeek := FFadingSeek;
  FCurrentFadingCross := FFadingCross;
end;

procedure TCrossfadingDirectSoundRenderer.SaveSettings;
var
  Reg: TRegistry;
begin
  Reg := TRegistry.Create;
  Reg.Rootkey := HKEY_CURRENT_USER;
  if Reg.OpenKey('SOFTWARE\DSP-worx\Crossfading Renderer', True) then
  begin
    Reg.WriteString('Device', GUIDToString(FGuid));
    Reg.WriteInteger('FadingSeek', FFadingSeek);
    Reg.WriteInteger('FadingCross', FFadingCross);
    Reg.WriteBool('AllowVideo', FAllowVideo);
    Reg.CloseKey;
  end;
  Reg.Free;
end;

function TCrossfadingDirectSoundRenderer.NonDelegatingQueryInterface(const IID: TGUID; out Obj): HResult;
begin
  if IsEqualGUID(IID, IID_IReferenceClock) or
     IsEqualGUID(IID, IID_IBasicAudio) or
     IsEqualGUID(IID, IID_IPersist) or
     IsEqualGUID(IID, IID_IMediaFilter) or
     IsEqualGUID(IID, IDispatch) or
     IsEqualGUID(IID, IID_IAMFilterMiscFlags) or
     IsEqualGUID(IID, IID_ISpecifyPropertyPages) or
     IsEqualGUID(IID, IID_ICrossfadingRenderer) then
  begin
    if GetInterface(IID, Obj)
      then Result := S_OK
      else Result := E_NOINTERFACE;
  end else
  begin
    Result := inherited NonDelegatingQueryInterface(IID, Obj);
  end;
end;

function TCrossfadingDirectSoundRenderer.EndOfStream: HResult;
begin
  FadeOut(False);
  LastWasEOS := True;
  Result := inherited EndOfStream;
end;

function TCrossfadingDirectSoundRenderer.Receive(MediaSample: IMediaSample): HResult;
var
  lStatus : DWORD;
  mi, mo : TReferenceTime;
  buffersize: Int64;
  lDelayMS : Int64;
begin
  if FPreBufferSize > FCurrentBufferSize then
  begin
    MediaSample.SetTime(nil,nil);
    inc(FCurrentBufferSize, MediaSample.GetActualDataLength);
    Result := inherited Receive(MediaSample);
    Exit;
  end;

  if MediaSample.GetTime(mi, mo) = S_OK then
  begin
    if FCurrentFadingSeek > FCurrentFadingCross
      then buffersize := FCurrentFadingSeek
      else buffersize := FCurrentFadingCross;
    lDelayMS := buffersize * 10000;
    mo := mo - lDelayMS;
    mi := mi - lDelayMS;
    MediaSample.SetTime(@mi,@mo);
  end;

  Result := inherited Receive(MediaSample);
  if FPreBufferSize <= FCurrentBufferSize then
  begin
    if not Assigned(FBuffer)
      then CreateDirectSoundBuffer(FMediaType);

    if Assigned(FBuffer) then
    begin
      FBuffer.GetStatus(lStatus);
      if lStatus and DSBSTATUS_PLAYING = 0 then
      begin
        if (FState = State_Running) and Assigned(FBuffer) then
        begin
          FadeIn;
        end;
      end;
    end;
  end;
end;

function TCrossfadingDirectSoundRenderer.BeginFlush: HResult;
begin
  FPos := 0;
  FCurrentBufferSize := 0;
  Result := inherited BeginFlush;
end;

function TCrossfadingDirectSoundRenderer.Run(StartTime: TReferenceTime): HResult;
begin
  Result := inherited Run(StartTime);
end;

procedure TCrossfadingDirectSoundRenderer.FadeIn;
begin
  if not assigned(FBuffer)
    then Exit;

  FBuffer.SetVolume(-10000);
  FBuffer._AddRef;
  SetRate(FRate);
  PostThreadMessage(FPlayThread.ThreadID, UM_PLAY, integer(FBuffer), 0);
  if FFirstRun and LastWasEOS
    then SetupFadeIn(FPlayThread.Device, FBuffer, FCurrentFadingCross, FVolume, FPlayThread.ThreadID)
    else SetupFadeIn(FPlayThread.Device, FBuffer, FCurrentFadingSeek, FVolume, FPlayThread.ThreadID);
  FFirstRun := False;
  LastWasEOS := False;
end;

procedure TCrossfadingDirectSoundRenderer.FadeOut(Seek: Boolean);
var
  s: Cardinal;
begin
  if not Assigned(FBuffer)
    then Exit;

  KillFading;
  FBuffer.GetStatus(s);
  if (s and DSBSTATUS_PLAYING = 0)
    then Exit;
  if not IsEndOfStream then
  begin
    if Seek
      then FLastFade := FCurrentFadingSeek
      else FLastFade := FCurrentFadingCross;
    SetupFadeOut(FPlayThread.Device, FBuffer, FLastFade);
  end else
  begin
    FLastFade := FCurrentFadingCross;
  end;
  FBuffer := nil;
  FPos := 0;
  FCurrentBufferSize := 0;
  CreateDirectSoundBuffer(FMediaType);
  if Seek then
  begin
    FGraph._AddRef;
    PostThreadMessage(FPlayThread.ThreadID, UM_SEEK, integer(FGraph), 0);
  end;
end;

procedure TCrossfadingDirectSoundRenderer.KillFading;
begin
  FLock.Lock;
  try
    PostThreadMessage(FPlayThread.ThreadID, UM_KILL, 0, 0);
  finally
    FLock.UnLock;
  end;
end;

function TCrossfadingDirectSoundRenderer.Stop: HResult;
begin
  Result := inherited Stop;
  FadeOut(True);
end;

function TCrossfadingDirectSoundRenderer.Pause: HResult;
begin
  Result := inherited Pause;
  FadeOut(True);
end;

function TCrossfadingDirectSoundRenderer.CheckMediaType(MediaType: PAMMediaType): HResult;
begin
  if not FAllowVideo and CheckVideoRenderer then
  begin
    Result := E_UNEXPECTED;
    Exit;
  end;

  if (MediaType = nil) then
  begin
    Result := E_POINTER;
    Exit;
  end;

  if not IsEqualGUID(MediaType.majortype, MEDIATYPE_Audio) or
     (not IsEqualGUID(MediaType.subtype, MEDIASUBTYPE_PCM) and
      not IsEqualGUID(MediaType.subtype, MEDIASUBTYPE_IEEE_FLOAT) and
      not IsEqualGUID(MediaType.subtype, MEDIASUBTYPE_DOLBY_AC3_SPDIF) and
      not IsEqualGUID(MediaType.subtype, MEDIASUBTYPE_RAW_SPORT) and
      not IsEqualGUID(MediaType.subtype, MEDIASUBTYPE_SPDIF_TAG_241h) and
      not IsEqualGUID(MediaType.subtype, MEDIASUBTYPE_DRM_Audio)) or
     not IsEqualGUID(MediaType.formattype, FORMAT_WaveFormatEx) then
  begin
    Result := E_INVALIDARG;
    Exit;
  end;

  Result := S_OK;
end;

function TCrossfadingDirectSoundRenderer.DoRenderSample(MediaSample: IMediaSample): HResult;
var
  pmt : PAMMediaType;
  Buffer : PByte;
  Size : Integer;
  p1, p2 : Pointer;
  s1, s2 : Cardinal;
begin
  FLock.Lock;
  try
    Result := S_OK;

    if MediaSample.GetMediaType(pmt) = S_OK then
    begin
      Result := CheckMediaType(pmt);
      if Result = S_OK then
      begin
        if Assigned(FMediaType)
          then DeleteMediaType(FMediaType);
        FMediaType := nil;

        FMediaType := CreateMediaType(pmt);

        Result := CreateDirectSoundBuffer(pmt);
      end;
      if Result <> DS_OK then
      begin
        Result := E_FAIL;
        Exit;
      end;
      FCurrentBufferSize := 0;
      FPos := 0;
      DeleteMediaType(pmt);
    end;

    MediaSample.GetPointer(Buffer);
    Size := MediaSample.GetActualDataLength;

    if Assigned(FBuffer) then
    begin
      // Lock the Buffer
      Result := FBuffer.Lock(FPos, Size, @p1, @s1, @p2, @s2, 0);

      // If the Buffer is lost, then Restore it and Lock it again
      if Result = DSERR_BUFFERLOST then
      begin
        FBuffer.Restore;
        FBuffer.Lock(FPos, Size, @p1, @s1, @p2, @s2, 0);
      end;

      // Copy the Buffer from the MediaSample to the Buffer(s)
      if (s1 > 0)
        then Move(Buffer^, p1^, s1);
      if (s2 > 0)
        then Move(PChar(Buffer)[s1], p2^, s2);

      FBuffer.Unlock(p1, s1, p2, s2);

      // store the next WriteCursor Position
      inc(FPos, Size);
      if FPos >= FBufferSize
        then dec(FPos,  FBufferSize);
    end;
  finally
    FLock.UnLock;
  end;

  Result := NOERROR;
end;

function TCrossfadingDirectSoundRenderer.SetMediaType(MediaType: PAMMediaType): HResult;
begin
  if (MediaType = nil) then
  begin
    Result := E_POINTER;
    Exit;
  end;

  if Assigned(FMediaType)
    then DeleteMediaType(FMediaType);
  FMediaType := nil;

  FMediaType := CreateMediaType(MediaType);

  Result := CreateDirectSoundBuffer(FMediaType);
end;

function TCrossfadingDirectSoundRenderer.CreateDirectSoundBuffer(MediaType : PAMMediaType) : HRESULT;
var
  lDesc : TDSBufferDesc;
  buffersize: Int64;
  s: Int64;
begin
  if FCurrentFadingSeek > FCurrentFadingCross
    then buffersize := FCurrentFadingSeek
    else buffersize := FCurrentFadingCross;

  fLock.Lock;
  try
    ZeroMemory(@lDesc,SizeOf(TDSBufferDesc));

    if Assigned(FBuffer) then
    begin
      FBuffer.Stop;
      FBuffer := nil;
    end;

    lDesc.dwSize := SizeOf(TDSBufferDesc);
    lDesc.dwFlags := DSBCAPS_GLOBALFOCUS or DSBCAPS_CTRLFREQUENCY or DSBCAPS_CTRLVOLUME or DSBCAPS_CTRLPAN or DSBCAPS_GETCURRENTPOSITION2;

    lDesc.lpwfxFormat := CoTaskMemAlloc(MediaType.cbFormat);
    Move(MediaType.pbFormat^, lDesc.lpwfxFormat^ ,MediaType.cbFormat);
    lDesc.guid3DAlgorithm := GUID_NULL;
    FBufferSize  := PWaveformatEx(MediaType.pbFormat)^.nSamplesPerSec *
                    PWaveformatEx(MediaType.pbFormat)^.nChannels *
                    (PWaveformatEx(MediaType.pbFormat)^.wBitsPerSample div 8) *
                    (buffersize + 1000) div 1000;

    lDesc.dwBufferBytes := FBufferSize;

    if FCurrentFadingSeek > FCurrentFadingCross
      then s := FCurrentFadingSeek
      else s := FCurrentFadingCross;

    FPreBufferSize := PWaveformatEx(MediaType.pbFormat)^.nAvgBytesPerSec * s div 1000;

    Result := FPlayThread.Device.CreateSoundBuffer(lDesc, FBuffer, nil);

    if (Result = DS_OK) and Assigned(FBuffer) then
    begin
      FBuffer.SetVolume(FVolume);
      FBuffer.SetPan(FBalance);
      FBuffer.GetFormat(@FFormat, SizeOf(FFormat), nil);
    end;
  finally
    fLock.UnLock;
  end;
end;

function TCrossfadingDirectSoundRenderer.GetDevice(out Guid: TGuid): HRESULT;
begin
  Guid := FGuid;
  Result := S_OK;
end;

function TCrossfadingDirectSoundRenderer.SetDevice(Guid: TGuid): HRESULT;
begin
  FGuid := Guid;
  Result := S_OK;
end;

function TCrossfadingDirectSoundRenderer.GetFadingSeekStop(out Fading: Integer): HRESULT;
begin
  Fading := FFadingSeek;
  Result := S_OK;
end;

function TCrossfadingDirectSoundRenderer.SetFadingSeekStop(Fading: Integer): HRESULT;
begin
  FFadingSeek := Fading;
  Result := S_OK;
end;

function TCrossfadingDirectSoundRenderer.GetFadingCross(out Fading: Integer): HRESULT;
begin
  Fading := FFadingCross;
  Result := S_OK;
end;

function TCrossfadingDirectSoundRenderer.SetFadingCross(Fading: Integer): HRESULT;
begin
  FFadingCross := Fading;
  Result := S_OK;
end;

function TCrossfadingDirectSoundRenderer.GetAllowVideo(out Allow: Boolean): HRESULT;
begin
  Allow := FAllowVideo;
  Result := S_OK;
end;

function TCrossfadingDirectSoundRenderer.SetAllowVideo(Allow: Boolean): HRESULT;
begin
  FAllowVideo := Allow;
  Result := S_OK;
end;

function TCrossfadingDirectSoundRenderer.CheckVideoRenderer: Boolean;
var
  Filter: IBaseFilter;
begin
  Result := Assigned(FGraph);
  if not Result
    then Exit;

  // ToDo: Find a way to detect if a Video Stream is present
  //       without using FGraph.EnumFilters().
  //       EnumFilters causes the Graph not to run ...

  try
    Result := FGraph.FindFilterByName('Video Mixing Renderer 9', Filter) = S_OK;
    if Result then Exit;
    Result := FGraph.FindFilterByName('Video Mixing Renderer 7', Filter) = S_OK;
    if Result then Exit;
    Result := FGraph.FindFilterByName('Video Renderer', Filter) = S_OK;
    if Result then Exit;
    Result := FGraph.FindFilterByName('Overlay Mixer', Filter) = S_OK;
    if Result then Exit;
    Result := FGraph.FindFilterByName('Overlay Mixer2', Filter) = S_OK;
    if Result then Exit;
  finally
    Filter := nil;
  end;
end;

function TCrossfadingDirectSoundRenderer.GetMediaPositionInterface(IID: TGUID; out Obj): HResult;
var
  hr: HResult;
begin
  FObjectCreationLock.Lock;
  try
    if Assigned(FPosition) then
    begin
      Result := FPosition.NonDelegatingQueryInterface(IID, Obj);
      Exit;
    end;

    hr := NOERROR;

    FPosition := TMyRendererPosPassThru.Create(Self, 'Renderer TBCPosPassThru', inherited GetOwner, hr, GetPin(0));
    if (FPosition = nil) then
    begin
      Result := E_OUTOFMEMORY;
      Exit;
    end;
    if (Failed(hr)) then
    begin
      FreeAndNil(FPosition);
      Result := E_NOINTERFACE;
      Exit;
    end;
    FPosition._AddRef;

    Result := GetMediaPositionInterface(IID, Obj);
  finally
    FObjectCreationLock.UnLock;
  end;
end;

procedure TCrossfadingDirectSoundRenderer.SetRate(dRate: Double);
begin
  FRate := dRate;
  if Assigned(FBuffer)
    then FBuffer.SetFrequency(Round(PWaveFormatEx(FMediaType.pbFormat).nSamplesPerSec * FRate))
end;

(*** IAMFilterMiscFlags methods ***********************************************)

function TCrossfadingDirectSoundRenderer.GetMiscFlags: ULONG;
begin
  Result := AM_FILTER_MISC_FLAGS_IS_RENDERER;
end;

(******************************************************************************)

initialization

  TBCClassFactory.CreateFilter(
    TCrossfadingDirectSoundRenderer, 'DC-Crossfading Renderer', CLSID_CrossfadingDirectSoundRenderer,
    CLSID_AudioRendererCategory, MERIT_PREFERRED + 1, 1, @SudPins
  );

end.


